#include "../include/PCLSegmentation.h"

//----------PlaneSegmentation-------------------
int PlaneSegmentation::segment (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
  pcl::PointIndices::Ptr &inliers, pcl::ModelCoefficients::Ptr &coefficients,
  double sac_distence_threshold, bool print_flag)
{
  // Create the segmentation object
  pcl::SACSegmentation<pcl::PointXYZ> seg;
  // Optional
  seg.setOptimizeCoefficients (true);
  // Mandatory
  seg.setModelType (pcl::SACMODEL_PLANE);
  seg.setMethodType (pcl::SAC_RANSAC);
  seg.setDistanceThreshold (sac_distence_threshold);

  seg.setInputCloud (cloud);
  seg.segment (*inliers, *coefficients);

  if (inliers->indices.size () == 0)
  {
    PCL_ERROR ("Could not estimate a planar model for the given dataset.");
    return (-1);
  }

  if (print_flag)
  {
    std::cerr << "Model coefficients: " << coefficients->values[0] << " "
                                        << coefficients->values[1] << " "
                                        << coefficients->values[2] << " "
                                        << coefficients->values[3] << std::endl;

    std::cerr << "Model inliers: " << inliers->indices.size () << std::endl;
    for (size_t i = 0; i < inliers->indices.size (); ++i)
      std::cerr << inliers->indices[i] << "    "  << cloud->points[inliers->indices[i]].x << " "
                                                  << cloud->points[inliers->indices[i]].y << " "
                                                  << cloud->points[inliers->indices[i]].z << std::endl;
  }
}

//--------------ExtractClusterOnPlane---------------------
void ExtractClusterOnPlane::extractCluster (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
  std::vector< pcl::PointCloud<pcl::PointXYZ> > &cloud_vector,
  double sac_distence_threshold, double cluster_tolerance,
  int min_cluster_size, int max_cluster_size, int sac_maxiterations)
{
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered (new pcl::PointCloud<pcl::PointXYZ>);

  pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);
  pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
  // Create the segmentation object
  pcl::SACSegmentation<pcl::PointXYZ> seg;
  // Optional
  seg.setOptimizeCoefficients (true);
  // Mandatory
  seg.setModelType (pcl::SACMODEL_PLANE);
  seg.setMethodType (pcl::SAC_RANSAC);
  seg.setMaxIterations (sac_maxiterations);
  seg.setDistanceThreshold (sac_distence_threshold);
  seg.setInputCloud (cloud);
  seg.segment (*inliers, *coefficients);

  // Extract the planar inliers from the input cloud
  pcl::ExtractIndices<pcl::PointXYZ> extract;
  extract.setInputCloud (cloud);
  extract.setIndices (inliers);

  // Remove the planar inliers, extract the rest, which are the clusters
  extract.setNegative (true);
  extract.filter (*cloud_filtered);
 // Creating the KdTree object for the search method of the extraction
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
  tree->setInputCloud (cloud_filtered);

  std::vector<pcl::PointIndices> cluster_indices;
  pcl::EuclideanClusterExtraction<pcl::PointXYZ> ec;
  ec.setClusterTolerance (cluster_tolerance); // meters
  ec.setMinClusterSize (min_cluster_size); //the min number of points
  ec.setMaxClusterSize (max_cluster_size);
  ec.setSearchMethod (tree);
  ec.setInputCloud (cloud_filtered);
  ec.extract (cluster_indices);

  inliers_.reset (new pcl::PointIndices);
  coefficients_.reset (new pcl::ModelCoefficients);
  inliers_ = inliers;
  coefficients_ =  coefficients;

  //get every cluster from cluster_indices
  int j = 0;
  for (std::vector<pcl::PointIndices>::const_iterator k = cluster_indices.begin (); k != cluster_indices.end (); ++k)
  {
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_cluster (new pcl::PointCloud<pcl::PointXYZ>);
    for (std::vector<int>::const_iterator pk = k->indices.begin (); pk != k->indices.end (); ++pk)
      cloud_cluster->points.push_back (cloud_filtered->points[*pk]);

    cloud_cluster->width = cloud_cluster->points.size ();
    cloud_cluster->height = 1;
    cloud_cluster->is_dense = true;

    cloud_vector.push_back (*cloud_cluster);

    std::cout << "Get the Cluster No."<< j <<" : "<<cloud_cluster->points.size () << " data points." << std::endl;

    j++;
  }
  cluster_number = j;
}

int ExtractClusterOnPlane::outputPlaneCoefficients (pcl::PointIndices::Ptr &inliers,
  pcl::ModelCoefficients::Ptr &coefficients, bool coefficients_print_flag)
{
  inliers = inliers_;
  coefficients = coefficients_;

  if (inliers->indices.size () == 0)
  {
    PCL_ERROR ("Could not estimate a planar model for the given dataset.");
    return (-1);
  }

  if (coefficients_print_flag)
  {
    std::cerr << "Model coefficients: " << coefficients->values[0] << " "
                                        << coefficients->values[1] << " "
                                        << coefficients->values[2] << " "
                                        << coefficients->values[3] << std::endl;
  }
}

//-------------RANSACPlane---------------------
void RANSACPlane::ransac (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
  pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_final,
  double distance_threshold)
{
  std::vector<int> inliers;
  // created RandomSampleConsensus object and compute the appropriated model
  pcl::SampleConsensusModelPlane<pcl::PointXYZ>::Ptr
  model_p (new pcl::SampleConsensusModelPlane<pcl::PointXYZ> (cloud));

  pcl::RandomSampleConsensus<pcl::PointXYZ> ransac (model_p);
  ransac.setDistanceThreshold (distance_threshold);
  ransac.computeModel();
  ransac.getInliers(inliers);

  // copies all inliers of the model computed to another PointCloud
  pcl::copyPointCloud<pcl::PointXYZ> (*cloud, inliers, *cloud_final);

}
